import importlib.metadata
import os
import tempfile

from authlib.integrations.flask_client import OAuth
from authlib.oidc.discovery import get_well_known_url
from cachelib.file import FileSystemCache
from cachelib.simple import SimpleCache
from flask import Flask
from flask import abort
from flask import current_app
from flask import g
from flask import has_request_context
from flask import redirect
from flask import request
from flask import session as flask_session
from flask import url_for
from flask_babel import gettext as _
from flask_session import Session

from auth_playground.endpoints import bp
from auth_playground.i18n import babel
from auth_playground.i18n import setup_i18n
from auth_playground.oauth import bp as oauth_bp
from auth_playground.session import ServerConfig

oauth = OAuth()
sess = Session()


def is_client_configured():
    """Check if OAuth client is configured by checking the registry."""
    return "default" in oauth._registry


def is_oauth_server_from_env(app):
    """Check if OAuth server is set via environment variables."""
    return bool(app.config.get("OAUTH_AUTH_SERVER"))


def is_server_configured():
    """Check if OAuth server is configured (from env or session)."""
    return is_oauth_server_from_env(current_app) or (
        g.server_config and g.server_config.issuer_url
    )


def is_oauth_client_from_env(app):
    """Check if OAuth client credentials are set via environment variables."""
    return bool(app.config.get("OAUTH_CLIENT_ID"))


def get_oauth_config(app):
    """Get OAuth config from environment or session, with env taking priority."""
    env_config = {
        "client_id": app.config.get("OAUTH_CLIENT_ID"),
        "client_secret": app.config.get("OAUTH_CLIENT_SECRET"),
        "auth_server": app.config.get("OAUTH_AUTH_SERVER"),
    }

    session_config = {}
    if has_request_context():
        session_config = flask_session.get("oauth_config", {})
    config = {
        **session_config,
        **{k: v for k, v in env_config.items() if v is not None},
    }

    return config


def setup_oauth(app):
    """Initialize OAuth client with configuration from environment or session."""
    if "default" in oauth._registry:
        del oauth._registry["default"]

    config = get_oauth_config(app)
    if not config:
        return False

    oauth.init_app(app)

    oauth.register(
        name="default",
        client_id=config["client_id"],
        client_secret=config["client_secret"],
        server_metadata_url=get_well_known_url(config["auth_server"], external=True),
        client_kwargs={"scope": "openid profile email phone address groups"},
    )
    return True


def setup_oauth_runtime(app, client_id, client_secret, auth_server):
    """Set up OAuth with runtime configuration stored in session."""
    flask_session["oauth_config"] = {
        "client_id": client_id,
        "client_secret": client_secret,
        "auth_server": auth_server,
    }

    return setup_oauth(app)


def create_app():
    """Create and configure the Flask application."""
    app = Flask(__name__)

    app.config["SECRET_KEY"] = os.environ.get(
        "SECRET_KEY", "dev-secret-key-change-in-production"
    )
    app.config["NAME"] = os.environ.get("APP_NAME", "Auth Playground")

    app.config["SESSION_TYPE"] = "cachelib"
    app.config["SESSION_CACHELIB"] = (
        SimpleCache()
        if app.debug
        else FileSystemCache(os.path.join(tempfile.gettempdir(), "flask_sessions"))
    )
    app.config["SESSION_PERMANENT"] = False

    app.config["BABEL_DEFAULT_LOCALE"] = "en"
    app.config["BABEL_DEFAULT_TIMEZONE"] = "UTC"

    oauth_auth_server_env = os.environ.get("OAUTH_AUTH_SERVER")
    app.config["OAUTH_CLIENT_ID"] = os.environ.get("OAUTH_CLIENT_ID")
    app.config["OAUTH_CLIENT_SECRET"] = os.environ.get("OAUTH_CLIENT_SECRET")
    app.config["OAUTH_AUTH_SERVER"] = (
        oauth_auth_server_env.rstrip("/") if oauth_auth_server_env else None
    )
    app.config["SERVER_NAME"] = os.environ.get("SERVER_NAME")

    sess.init_app(app)
    setup_i18n(app)

    app.register_blueprint(oauth_bp)
    app.register_blueprint(bp, url_prefix="/<lang:lang_code>")

    @app.route("/")
    def root():
        """Redirect to the appropriate language prefix based on browser preference."""
        available_langs = [locale.language for locale in babel.list_translations()]
        default_lang = app.config.get("BABEL_DEFAULT_LOCALE", "en")
        lang = request.accept_languages.best_match(available_langs) or default_lang
        return redirect(url_for("routes.index", lang_code=lang))

    @app.before_request
    def redirect_to_server_name():
        """Redirect requests to SERVER_NAME if host doesn't match."""
        server_name = app.config.get("SERVER_NAME")
        if not server_name or request.host == server_name:
            return None

        if request.method != "GET":
            abort(
                400,
                _("This request should be sent to %(server_name)s")
                % {"server_name": server_name},
            )

        url = f"{request.scheme}://{server_name}{request.path}"
        if request.query_string:
            url += "?" + request.query_string.decode()

        return redirect(url)

    @app.before_request
    def load_server_config():
        """Load server configuration into flask.g."""
        setup_oauth(app)

        if not g.get("server_config"):
            g.server_config = ServerConfig.deserialize(flask_session)
            if not g.server_config:
                g.server_config = ServerConfig()
            if issuer_url := app.config.get("OAUTH_AUTH_SERVER"):
                g.server_config.issuer_url = issuer_url

    @app.after_request
    def save_server_config(response):
        """Automatically save server configuration to session after each request."""
        if (
            g.get("server_config")
            and not app.config.get("OAUTH_AUTH_SERVER")
            and (g.server_config.metadata or g.server_config.issuer_url)
        ):
            g.server_config.save(flask_session)
        return response

    @app.context_processor
    def inject_server_info():
        """Inject server information into all templates."""
        pkg_metadata = importlib.metadata.metadata("auth-playground")
        project_urls = dict(
            [url.split(", ", 1) for url in pkg_metadata.get_all("Project-URL") or []]
        )
        return {
            "app_version": importlib.metadata.version("auth-playground"),
            "repository_url": project_urls.get("repository"),
        }

    return app



class ServerSpecs:
    """Detects which OAuth 2.0 and OpenID Connect specifications a server implements."""

    def __init__(self, server_metadata: dict | None = None):
        """Initialize ServerSpecs with server metadata."""
        self._metadata = server_metadata or {}

    @property
    def oauth_2_authorization_framework(self) -> bool:
        """RFC 6749 - OAuth 2.0 Authorization Framework."""
        return "authorization_endpoint" in self._metadata

    @property
    def oauth_2_bearer_token_usage(self) -> bool:
        """RFC 6750 - OAuth 2.0 Bearer Token Usage."""
        return self.oauth_2_authorization_framework

    @property
    def oidc_core(self) -> bool:
        """OpenID Connect Core 1.0."""
        return all(
            key in self._metadata
            for key in [
                "issuer",
                "jwks_uri",
                "id_token_signing_alg_values_supported",
                "response_types_supported",
                "subject_types_supported",
            ]
        )

    @property
    def oidc_discovery(self) -> bool:
        """OpenID Connect Discovery 1.0."""
        return self.oidc_core

    @property
    def oauth_2_token_revocation(self) -> bool:
        """RFC 7009 - OAuth 2.0 Token Revocation."""
        return "revocation_endpoint" in self._metadata

    @property
    def oauth_2_token_introspection(self) -> bool:
        """RFC 7662 - OAuth 2.0 Token Introspection."""
        return "introspection_endpoint" in self._metadata

    @property
    def oauth_2_dynamic_client_registration(self) -> bool:
        """RFC 7591 - OAuth 2.0 Dynamic Client Registration."""
        return "registration_endpoint" in self._metadata

    @property
    def oauth_2_dynamic_client_registration_management(self) -> bool:
        """RFC 7592 - OAuth 2.0 Dynamic Client Registration Management."""
        return self.oauth_2_dynamic_client_registration

    @property
    def oidc_registration(self) -> bool:
        """OpenID Connect Dynamic Client Registration 1.0."""
        return self.oauth_2_dynamic_client_registration and self.oidc_core

    @property
    def oauth_2_pkce(self) -> bool:
        """RFC 7636 - Proof Key for Code Exchange (PKCE)."""
        return "code_challenge_methods_supported" in self._metadata

    @property
    def oauth_2_pushed_authorization_requests(self) -> bool:
        """RFC 9126 - OAuth 2.0 Pushed Authorization Requests (PAR)."""
        return "pushed_authorization_request_endpoint" in self._metadata

    @property
    def oauth_2_jwt_secured_authorization_request(self) -> bool:
        """RFC 9101 - JWT-Secured Authorization Request (JAR)."""
        return (
            "require_signed_request_object" in self._metadata
            or "request_object_signing_alg_values_supported" in self._metadata
        )

    @property
    def oauth_2_authorization_server_issuer_identification(self) -> bool:
        """RFC 9207 - OAuth 2.0 Authorization Server Issuer Identification."""
        return "authorization_response_iss_parameter_supported" in self._metadata

    @property
    def oauth_2_step_up_authentication_challenge(self) -> bool:
        """RFC 9470 - OAuth 2.0 Step Up Authentication Challenge Protocol."""
        return "acr_values_supported" in self._metadata

    @property
    def oauth_2_authorization_server_metadata(self) -> bool:
        """RFC 8414 - OAuth 2.0 Authorization Server Metadata."""
        return "issuer" in self._metadata

    @property
    def oauth_2_jwt_access_tokens(self) -> bool | None:
        """RFC 9068 - JSON Web Token (JWT) Profile for OAuth 2.0 Access Tokens."""
        return None

    @property
    def json_web_token(self) -> bool:
        """RFC 7519 - JSON Web Token (JWT)."""
        return self.oidc_core

    @property
    def json_web_signature(self) -> bool:
        """RFC 7515 - JSON Web Signature (JWS)."""
        return self.oidc_core

    @property
    def json_web_encryption(self) -> bool | None:
        """RFC 7516 - JSON Web Encryption (JWE)."""
        return None

    @property
    def json_web_key(self) -> bool:
        """RFC 7517 - JSON Web Key (JWK)."""
        return "jwks_uri" in self._metadata

    @property
    def json_web_algorithms(self) -> bool | None:
        """RFC 7518 - JSON Web Algorithms (JWA)."""
        return None

    @property
    def oauth_2_device_authorization_grant(self) -> bool:
        """RFC 8628 - OAuth 2.0 Device Authorization Grant."""
        return "device_authorization_endpoint" in self._metadata

    @property
    def oauth_2_token_exchange(self) -> bool | None:
        """RFC 8693 - OAuth 2.0 Token Exchange."""
        return None

    @property
    def oidc_ciba(self) -> bool:
        """OpenID Connect Client-Initiated Backchannel Authentication Flow."""
        return "backchannel_authentication_endpoint" in self._metadata

    @property
    def oauth_2_mtls(self) -> bool:
        """RFC 8705 - OAuth 2.0 Mutual-TLS Client Authentication and Certificate-Bound Access Tokens."""
        return (
            "tls_client_certificate_bound_access_tokens" in self._metadata
            or "mtls_endpoint_aliases" in self._metadata
        )

    @property
    def oauth_2_dpop(self) -> bool:
        """RFC 9449 - OAuth 2.0 Demonstrating Proof of Possession (DPoP)."""
        return "dpop_signing_alg_values_supported" in self._metadata

    @property
    def oauth_2_rich_authorization_requests(self) -> bool:
        """RFC 9396 - OAuth 2.0 Rich Authorization Requests."""
        return "authorization_details_types_supported" in self._metadata

    @property
    def oauth_2_multiple_response_types(self) -> bool:
        """OAuth 2.0 Multiple Response Types."""
        return "response_modes_supported" in self._metadata

    @property
    def oauth_2_form_post_response_mode(self) -> bool:
        """OAuth 2.0 Form Post Response Mode."""
        response_modes = self._metadata.get("response_modes_supported", [])
        return "form_post" in response_modes

    @property
    def oauth_2_jarm(self) -> bool:
        """JWT Secured Authorization Response Mode for OAuth 2.0 (JARM)."""
        response_modes = self._metadata.get("response_modes_supported", [])
        return "authorization_signing_alg_values_supported" in self._metadata or any(
            mode.endswith(".jwt") or mode == "jwt" for mode in response_modes
        )

    @property
    def oidc_session_management(self) -> bool:
        """OpenID Connect Session Management 1.0."""
        return "check_session_iframe" in self._metadata

    @property
    def oidc_frontchannel_logout(self) -> bool:
        """OpenID Connect Front-Channel Logout 1.0."""
        return "frontchannel_logout_supported" in self._metadata

    @property
    def oidc_backchannel_logout(self) -> bool:
        """OpenID Connect Back-Channel Logout 1.0."""
        return "backchannel_logout_supported" in self._metadata

    @property
    def oidc_rpinitiated_logout(self) -> bool:
        """OpenID Connect RP-Initiated Logout 1.0."""
        return "end_session_endpoint" in self._metadata

    @property
    def oidc_prompt_create(self) -> bool:
        """OpenID Connect Prompt Create 1.0."""
        prompt_values = self._metadata.get("prompt_values_supported", [])
        return "create" in prompt_values

    @property
    def fapi_1_baseline(self) -> bool:
        """Financial-grade API 1.0 - Part 1: Baseline."""
        if (
            self.oauth_2_pushed_authorization_requests
            and self.oauth_2_jwt_secured_authorization_request
            and self.oauth_2_jarm
        ):
            return True
        elif self.oauth_2_pkce and self.oidc_core:
            return True
        return False

    @property
    def fapi_1_advanced(self) -> bool:
        """Financial-grade API 1.0 - Part 2: Advanced."""
        return (
            self.oauth_2_pushed_authorization_requests
            and self.oauth_2_jwt_secured_authorization_request
            and self.oauth_2_jarm
        )

    @property
    def openid_for_verifiable_credential_issuance(self) -> bool:
        """OpenID for Verifiable Credential Issuance."""
        return (
            "credential_issuer" in self._metadata
            or "credential_endpoint" in self._metadata
        )

    @property
    def openid_for_verifiable_presentations(self) -> bool:
        """OpenID for Verifiable Presentations."""
        return "vp_formats_supported" in self._metadata

    @property
    def oauth_2_resource_indicators(self) -> bool:
        """RFC 8707 - Resource Indicators for OAuth 2.0."""
        return "resource_signing_alg_values_supported" in self._metadata or bool(
            self._metadata.get("resource")
        )

    @property
    def oauth_2_assertion_framework(self) -> bool | None:
        """RFC 7521 - Assertion Framework for OAuth 2.0 Client Authentication and Authorization Grants."""
        return None

    @property
    def oauth_2_saml_profile(self) -> bool | None:
        """RFC 7522 - Security Assertion Markup Language (SAML) 2.0 Profile for OAuth 2.0."""
        return None

    @property
    def oauth_2_jwt_profile(self) -> bool | None:
        """RFC 7523 - JSON Web Token (JWT) Profile for OAuth 2.0 Client Authentication and Authorization Grants."""
        return None

    @property
    def jwt_proof_of_possession(self) -> bool | None:
        """RFC 7800 - Proof-of-Possession Key Semantics for JSON Web Tokens (JWTs)."""
        return None

    @property
    def jwt_best_current_practices(self) -> bool | None:
        """RFC 8725 - JSON Web Token Best Current Practices."""
        return None

    @property
    def oauth_2_security_best_current_practice(self) -> bool | None:
        """RFC 9700 - OAuth 2.0 Security Best Current Practice."""
        return None

    @property
    def jwk_thumbprint_uri(self) -> bool | None:
        """RFC 9278 - JWK Thumbprint URI."""
        return None

    @property
    def oauth_2_jwt_introspection_response(self) -> bool | None:
        """RFC 9701 - JWT Response for OAuth Token Introspection."""
        return None

    @property
    def ietf_urn_sub_namespace_oauth(self) -> bool | None:
        """RFC 6755 - An IETF URN Sub-Namespace for OAuth."""
        return None

    @property
    def oauth_2_threat_model_security_considerations(self) -> bool | None:
        """RFC 6819 - OAuth 2.0 Threat Model and Security Considerations."""
        return None

    @property
    def oauth_2_native_apps(self) -> bool:
        """RFC 8252 - OAuth 2.0 for Native Apps."""
        return self._metadata.get("require_pushed_authorization_requests", False) or (
            self.oauth_2_pkce and self.oauth_2_pushed_authorization_requests
        )

    def _get_all_property_names(self) -> list[str]:
        """Get a list of all property names."""
        return [
            name
            for name in dir(self)
            if isinstance(getattr(type(self), name, None), property)
        ]

    def to_dict(self) -> dict[str, bool | None]:
        """Convert the specs state to a dictionary."""
        return {
            field_name: getattr(self, field_name)
            for field_name in self._get_all_property_names()
        }

    def get_supported_specs(self) -> list[str]:
        """Get a list of supported specification names."""
        return [
            field_name
            for field_name, supported in self.to_dict().items()
            if supported is True
        ]

    def get_unsupported_specs(self) -> list[str]:
        """Get a list of unsupported specification names."""
        return [
            field_name
            for field_name, supported in self.to_dict().items()
            if supported is False
        ]

    def get_unknown_specs(self) -> list[str]:
        """Get a list of specifications with unknown support."""
        return [
            field_name
            for field_name, supported in self.to_dict().items()
            if supported is None
        ]

    def get_spec_url(self, field_name: str) -> str:
        """Get the URL for a specification."""
        urls = {
            "oauth_2_authorization_framework": "https://datatracker.ietf.org/doc/html/rfc6749",
            "oauth_2_bearer_token_usage": "https://datatracker.ietf.org/doc/html/rfc6750",
            "ietf_urn_sub_namespace_oauth": "https://datatracker.ietf.org/doc/html/rfc6755",
            "oauth_2_threat_model_security_considerations": "https://datatracker.ietf.org/doc/html/rfc6819",
            "oauth_2_token_revocation": "https://datatracker.ietf.org/doc/html/rfc7009",
            "json_web_signature": "https://datatracker.ietf.org/doc/html/rfc7515",
            "json_web_encryption": "https://datatracker.ietf.org/doc/html/rfc7516",
            "json_web_key": "https://datatracker.ietf.org/doc/html/rfc7517",
            "json_web_algorithms": "https://datatracker.ietf.org/doc/html/rfc7518",
            "json_web_token": "https://datatracker.ietf.org/doc/html/rfc7519",
            "oauth_2_assertion_framework": "https://datatracker.ietf.org/doc/html/rfc7521",
            "oauth_2_saml_profile": "https://datatracker.ietf.org/doc/html/rfc7522",
            "oauth_2_jwt_profile": "https://datatracker.ietf.org/doc/html/rfc7523",
            "oauth_2_dynamic_client_registration": "https://datatracker.ietf.org/doc/html/rfc7591",
            "oauth_2_dynamic_client_registration_management": "https://datatracker.ietf.org/doc/html/rfc7592",
            "oauth_2_pkce": "https://datatracker.ietf.org/doc/html/rfc7636",
            "oauth_2_token_introspection": "https://datatracker.ietf.org/doc/html/rfc7662",
            "jwt_proof_of_possession": "https://datatracker.ietf.org/doc/html/rfc7800",
            "oauth_2_native_apps": "https://datatracker.ietf.org/doc/html/rfc8252",
            "oauth_2_authorization_server_metadata": "https://datatracker.ietf.org/doc/html/rfc8414",
            "oauth_2_device_authorization_grant": "https://datatracker.ietf.org/doc/html/rfc8628",
            "oauth_2_token_exchange": "https://datatracker.ietf.org/doc/html/rfc8693",
            "oauth_2_mtls": "https://datatracker.ietf.org/doc/html/rfc8705",
            "oauth_2_resource_indicators": "https://datatracker.ietf.org/doc/html/rfc8707",
            "jwt_best_current_practices": "https://datatracker.ietf.org/doc/html/rfc8725",
            "oauth_2_jwt_access_tokens": "https://datatracker.ietf.org/doc/html/rfc9068",
            "oauth_2_jwt_secured_authorization_request": "https://datatracker.ietf.org/doc/html/rfc9101",
            "oauth_2_pushed_authorization_requests": "https://datatracker.ietf.org/doc/html/rfc9126",
            "oauth_2_authorization_server_issuer_identification": "https://datatracker.ietf.org/doc/html/rfc9207",
            "jwk_thumbprint_uri": "https://datatracker.ietf.org/doc/html/rfc9278",
            "oauth_2_rich_authorization_requests": "https://datatracker.ietf.org/doc/html/rfc9396",
            "oauth_2_dpop": "https://datatracker.ietf.org/doc/html/rfc9449",
            "oauth_2_step_up_authentication_challenge": "https://datatracker.ietf.org/doc/html/rfc9470",
            "oauth_2_security_best_current_practice": "https://datatracker.ietf.org/doc/html/rfc9700",
            "oauth_2_jwt_introspection_response": "https://datatracker.ietf.org/doc/html/rfc9701",
            "oidc_core": "https://openid.net/specs/openid-connect-core-1_0.html",
            "oidc_discovery": "https://openid.net/specs/openid-connect-discovery-1_0.html",
            "oidc_registration": "https://openid.net/specs/openid-connect-registration-1_0.html",
            "oidc_session_management": "https://openid.net/specs/openid-connect-session-1_0.html",
            "oidc_frontchannel_logout": "https://openid.net/specs/openid-connect-frontchannel-1_0.html",
            "oidc_backchannel_logout": "https://openid.net/specs/openid-connect-backchannel-1_0.html",
            "oidc_rpinitiated_logout": "https://openid.net/specs/openid-connect-rpinitiated-1_0.html",
            "oidc_prompt_create": "https://openid.net/specs/openid-connect-prompt-create-1_0.html",
            "oauth_2_multiple_response_types": "https://openid.net/specs/oauth-v2-multiple-response-types-1_0.html",
            "oauth_2_form_post_response_mode": "https://openid.net/specs/oauth-v2-form-post-response-mode-1_0.html",
            "oauth_2_jarm": "https://openid.net/specs/oauth-v2-jarm.html",
            "oidc_ciba": "https://openid.net/specs/openid-client-initiated-backchannel-authentication-core-1_0.html",
            "fapi_1_baseline": "https://openid.net/specs/openid-financial-api-part-1-1_0.html",
            "fapi_1_advanced": "https://openid.net/specs/openid-financial-api-part-2-1_0.html",
            "openid_for_verifiable_credential_issuance": "https://openid.net/specs/openid-4-verifiable-credential-issuance-1_0.html",
            "openid_for_verifiable_presentations": "https://openid.net/specs/openid-4-verifiable-presentations-1_0.html",
        }
        return urls.get(field_name, "#")

    def get_spec_display_name(self, field_name: str) -> str:
        """Get a human-readable display name for a specification."""
        display_names = {
            "oauth_2_authorization_framework": "RFC 6749 - OAuth 2.0 Authorization Framework",
            "oauth_2_bearer_token_usage": "RFC 6750 - OAuth 2.0 Bearer Token Usage",
            "ietf_urn_sub_namespace_oauth": "RFC 6755 - IETF URN Sub-Namespace for OAuth",
            "oauth_2_threat_model_security_considerations": "RFC 6819 - OAuth 2.0 Threat Model and Security Considerations",
            "oauth_2_token_revocation": "RFC 7009 - OAuth 2.0 Token Revocation",
            "json_web_signature": "RFC 7515 - JSON Web Signature (JWS)",
            "json_web_encryption": "RFC 7516 - JSON Web Encryption (JWE)",
            "json_web_key": "RFC 7517 - JSON Web Key (JWK)",
            "json_web_algorithms": "RFC 7518 - JSON Web Algorithms (JWA)",
            "json_web_token": "RFC 7519 - JSON Web Token (JWT)",
            "oauth_2_assertion_framework": "RFC 7521 - Assertion Framework for OAuth 2.0",
            "oauth_2_saml_profile": "RFC 7522 - SAML 2.0 Profile for OAuth 2.0",
            "oauth_2_jwt_profile": "RFC 7523 - JWT Profile for OAuth 2.0",
            "oauth_2_dynamic_client_registration": "RFC 7591 - OAuth 2.0 Dynamic Client Registration",
            "oauth_2_dynamic_client_registration_management": "RFC 7592 - OAuth 2.0 Dynamic Client Registration Management",
            "oauth_2_pkce": "RFC 7636 - PKCE",
            "oauth_2_token_introspection": "RFC 7662 - OAuth 2.0 Token Introspection",
            "jwt_proof_of_possession": "RFC 7800 - Proof-of-Possession Key Semantics for JWTs",
            "oauth_2_native_apps": "RFC 8252 - OAuth 2.0 for Native Apps",
            "oauth_2_authorization_server_metadata": "RFC 8414 - OAuth 2.0 Authorization Server Metadata",
            "oauth_2_device_authorization_grant": "RFC 8628 - Device Authorization Grant",
            "oauth_2_token_exchange": "RFC 8693 - OAuth 2.0 Token Exchange",
            "oauth_2_mtls": "RFC 8705 - OAuth 2.0 Mutual-TLS",
            "oauth_2_resource_indicators": "RFC 8707 - Resource Indicators for OAuth 2.0",
            "jwt_best_current_practices": "RFC 8725 - JWT Best Current Practices",
            "oauth_2_jwt_access_tokens": "RFC 9068 - JWT Profile for OAuth 2.0 Access Tokens",
            "oauth_2_jwt_secured_authorization_request": "RFC 9101 - JWT-Secured Authorization Request (JAR)",
            "oauth_2_pushed_authorization_requests": "RFC 9126 - Pushed Authorization Requests (PAR)",
            "oauth_2_authorization_server_issuer_identification": "RFC 9207 - Authorization Server Issuer Identification",
            "jwk_thumbprint_uri": "RFC 9278 - JWK Thumbprint URI",
            "oauth_2_rich_authorization_requests": "RFC 9396 - Rich Authorization Requests (RAR)",
            "oauth_2_dpop": "RFC 9449 - Demonstrating Proof of Possession (DPoP)",
            "oauth_2_step_up_authentication_challenge": "RFC 9470 - Step Up Authentication Challenge",
            "oauth_2_security_best_current_practice": "RFC 9700 - OAuth 2.0 Security Best Current Practice",
            "oauth_2_jwt_introspection_response": "RFC 9701 - JWT Response for OAuth Token Introspection",
            "oidc_core": "OpenID Connect Core 1.0",
            "oidc_discovery": "OpenID Connect Discovery 1.0",
            "oidc_registration": "OpenID Connect Dynamic Registration 1.0",
            "oidc_session_management": "OpenID Connect Session Management 1.0",
            "oidc_frontchannel_logout": "OpenID Connect Front-Channel Logout 1.0",
            "oidc_backchannel_logout": "OpenID Connect Back-Channel Logout 1.0",
            "oidc_rpinitiated_logout": "OpenID Connect RP-Initiated Logout 1.0",
            "oidc_prompt_create": "OpenID Connect Prompt Create 1.0",
            "oauth_2_multiple_response_types": "OAuth 2.0 Multiple Response Types",
            "oauth_2_form_post_response_mode": "OAuth 2.0 Form Post Response Mode",
            "oauth_2_jarm": "JWT Secured Authorization Response Mode (JARM)",
            "oidc_ciba": "OpenID Connect CIBA",
            "fapi_1_baseline": "FAPI 1.0 Baseline",
            "fapi_1_advanced": "FAPI 1.0 Advanced",
            "openid_for_verifiable_credential_issuance": "OpenID for Verifiable Credential Issuance",
            "openid_for_verifiable_presentations": "OpenID for Verifiable Presentations",
        }
        return display_names.get(field_name, field_name.replace("_", " ").title())



import importlib.metadata
import uuid

import requests
from authlib.common.errors import AuthlibBaseError
from authlib.common.urls import add_params_to_uri
from flask import Blueprint
from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import request
from flask import session
from flask import url_for
from flask_babel import gettext as _

import auth_playground
from auth_playground.decorators import server_config_needed
from auth_playground.forms import AuthorizationParamsForm
from auth_playground.forms import DynamicRegistrationForm
from auth_playground.forms import RefreshTokenForm
from auth_playground.forms import UnregisterClientForm

bp = Blueprint("oauth", __name__)


def get_software_id() -> str:
    """Get unique software identifier based on repository URL."""
    pkg_metadata = importlib.metadata.metadata("auth-playground")
    project_urls = dict(
        [url.split(", ", 1) for url in pkg_metadata.get_all("Project-URL") or []]
    )
    repository_url = project_urls.get("repository")
    return str(
        uuid.uuid5(uuid.NAMESPACE_URL, repository_url)
        if repository_url
        else uuid.uuid4()
    )


def get_software_version() -> str:
    """Get software version from package metadata."""
    return importlib.metadata.version("auth-playground")


def clear_user_session():
    """Clear user and token data from the session."""
    try:
        del session["user"]
    except KeyError:
        pass

    try:
        del session["token"]
    except KeyError:
        pass


@bp.route("/client/dynamic-registration", methods=["POST"])
@server_config_needed(client_needed=False)
def client_dynamic_registration():
    """Automatically register OAuth client using dynamic client registration."""
    form = DynamicRegistrationForm()

    if not form.validate_on_submit():
        flash(_("Invalid request"), "error")
        return redirect(url_for("routes.configure_client"))

    if not g.server_config.specs.oauth_2_dynamic_client_registration:
        flash(_("Dynamic client registration not supported"), "error")
        return redirect(url_for("routes.configure_client"))

    registration_endpoint = g.server_config.metadata["registration_endpoint"]

    redirect_uris = [
        url_for("oauth.authorize_callback", _external=True),
    ]

    registration_data = {
        "client_name": "Auth Playground",
        "client_uri": url_for("routes.index", _external=True),
        "redirect_uris": redirect_uris,
        "grant_types": ["authorization_code", "refresh_token"],
        "response_types": ["code"],
        "token_endpoint_auth_method": "client_secret_basic",
        "scope": "openid profile email phone address groups",
        "tos_uri": url_for("routes.tos", _external=True),
        "policy_uri": url_for("routes.policy", _external=True),
        "software_id": get_software_id(),
        "software_version": get_software_version(),
    }

    if g.server_config.specs.oidc_rpinitiated_logout:
        registration_data["post_logout_redirect_uris"] = [
            url_for("oauth.logout_callback", _external=True),
        ]

    headers = {}
    initial_access_token = form.initial_access_token.data
    if initial_access_token:
        headers["Authorization"] = f"Bearer {initial_access_token}"

    try:
        response = requests.post(
            registration_endpoint, json=registration_data, headers=headers, timeout=10
        )
        response.raise_for_status()
        client_data = response.json()
    except requests.RequestException as e:
        error_message = _("Dynamic client registration failed")
        if hasattr(e, "response") and e.response is not None:
            try:
                error_data = e.response.json()
                if "error_description" in error_data:
                    error_message = _(
                        "Dynamic client registration failed: {error}"
                    ).format(error=error_data["error_description"])
                elif "error" in error_data:
                    error_message = _(
                        "Dynamic client registration failed: {error}"
                    ).format(error=error_data["error"])
            except ValueError:
                pass  # Response is not JSON, use default message
        flash(error_message, "error")
        return redirect(url_for("routes.configure_client"))
    except ValueError:
        flash(_("Invalid JSON response from registration endpoint"), "error")
        return redirect(url_for("routes.configure_client"))

    auth_playground.setup_oauth_runtime(
        current_app,
        client_data["client_id"],
        client_data["client_secret"],
        g.server_config.issuer_url,
    )

    if "registration_access_token" in client_data:
        g.server_config.registration_access_token = client_data[
            "registration_access_token"
        ]
    if "registration_client_uri" in client_data:
        g.server_config.registration_client_uri = client_data["registration_client_uri"]

    flash(
        _("Client successfully registered!"),
        "success",
    )
    return redirect(url_for("routes.playground"))


@bp.route("/unregister-client", methods=["POST"])
@server_config_needed
def unregister_client():
    """Unregister OAuth client using dynamic client registration management."""
    form = UnregisterClientForm()

    if not form.validate_on_submit():
        flash(_("Invalid request"), "error")
        return redirect(url_for("routes.playground"))

    if (
        not g.server_config.registration_access_token
        or not g.server_config.registration_client_uri
    ):
        flash(_("Client registration management credentials not found"), "error")
        return redirect(url_for("routes.playground"))

    headers = {"Authorization": f"Bearer {g.server_config.registration_access_token}"}

    try:
        response = requests.delete(
            g.server_config.registration_client_uri, headers=headers, timeout=10
        )
        response.raise_for_status()
    except requests.RequestException as e:
        error_message = _("Client unregistration failed")
        if hasattr(e, "response") and e.response is not None:
            try:
                error_data = e.response.json()
                if "error_description" in error_data:
                    error_message = _("Client unregistration failed: {error}").format(
                        error=error_data["error_description"]
                    )
                elif "error" in error_data:
                    error_message = _("Client unregistration failed: {error}").format(
                        error=error_data["error"]
                    )
            except ValueError:
                pass  # Response is not JSON, use default message
        flash(error_message, "error")
        return redirect(url_for("routes.playground"))

    session.pop("oauth_config", None)
    session.pop("user", None)
    session.pop("token", None)

    g.server_config.registration_access_token = None
    g.server_config.registration_client_uri = None

    flash(_("Client successfully unregistered"), "success")
    return redirect(url_for("routes.configure_client"))


@bp.route("/authorize", methods=["GET", "POST"])
@server_config_needed
def authorize():
    """Redirect users to the Identity Provider authorization page."""
    kwargs = {}

    if request.method == "POST":
        form = AuthorizationParamsForm()

        if form.validate_on_submit():
            if form.scopes.data:
                kwargs["scope"] = " ".join(form.scopes.data)

            if form.prompt.data:
                kwargs["prompt"] = form.prompt.data

            if form.ui_locales.data:
                kwargs["ui_locales"] = form.ui_locales.data

    if request.method == "GET" and "user" in session:
        kwargs["prompt"] = "login"

    return auth_playground.oauth.default.authorize_redirect(
        url_for("oauth.authorize_callback", _external=True), **kwargs
    )


@bp.route("/authorize_callback")
def authorize_callback():
    """Handle OAuth callback after user authorization."""
    try:
        token = auth_playground.oauth.default.authorize_access_token()
        session["user"] = token.get("userinfo")
        session["token"] = {
            "access_token": token.get("access_token"),
            "refresh_token": token.get("refresh_token"),
            "id_token": token.get("id_token"),
            "token_type": token.get("token_type"),
            "expires_in": token.get("expires_in"),
            "expires_at": token.get("expires_at"),
            "scope": token.get("scope"),
        }
        flash(_("You have been successfully logged in."), "success")
    except AuthlibBaseError as exc:
        flash(
            _("An error happened during login: {error}").format(error=exc.description),
            "error",
        )

    return redirect(url_for("routes.tokens"))


@bp.route("/logout/local")
def logout_local():
    """Log out locally without contacting the Identity Provider."""
    clear_user_session()
    flash(_("You have been logged out"), "success")
    return redirect(url_for("routes.playground"))


@bp.route("/logout")
@server_config_needed
def logout():
    """Redirect users to the Identity Provider logout page for global logout."""
    auth_playground.oauth.default.load_server_metadata()
    end_session_endpoint = auth_playground.oauth.default.server_metadata.get(
        "end_session_endpoint"
    )
    id_token = session.get("token", {}).get("id_token")

    oauth_config = auth_playground.get_oauth_config(current_app)
    client_id = oauth_config["client_id"] if oauth_config else None

    end_session_url = add_params_to_uri(
        end_session_endpoint,
        dict(
            client_id=client_id,
            id_token_hint=id_token,
            post_logout_redirect_uri=url_for("oauth.logout_callback", _external=True),
        ),
    )
    return redirect(end_session_url)


@bp.route("/logout_callback")
def logout_callback():
    """Handle callback after server-side logout."""
    clear_user_session()
    flash(_("You have been logged out from the server"), "success")
    return redirect(url_for("routes.playground"))


@bp.route("/refresh", methods=["POST"])
def refresh():
    """Refresh the access token using the refresh token."""
    form = RefreshTokenForm()
    if not form.validate_on_submit():
        flash(_("Invalid request"), "error")
        return redirect(url_for("routes.playground"))

    refresh_token = session.get("token", {}).get("refresh_token")
    if not refresh_token:
        flash(_("No refresh token available"), "error")
        return redirect(url_for("routes.playground"))

    try:
        original_scope = session.get("token", {}).get("scope", "")
        new_token = auth_playground.oauth.default.fetch_access_token(
            grant_type="refresh_token",
            refresh_token=refresh_token,
            scope=original_scope,
        )

        old_token = session.get("token", {})
        session["token"] = {
            "access_token": new_token.get("access_token"),
            "refresh_token": new_token.get("refresh_token") or refresh_token,
            "id_token": new_token.get("id_token") or old_token.get("id_token"),
            "token_type": new_token.get("token_type"),
            "expires_in": new_token.get("expires_in"),
            "expires_at": new_token.get("expires_at"),
            "scope": new_token.get("scope"),
        }
        flash(_("Token successfully refreshed"), "success")
    except AuthlibBaseError as exc:
        flash(
            _("An error happened during token refresh: {error}").format(
                error=exc.description
            ),
            "error",
        )

    return redirect(url_for("routes.playground"))



from flask import g
from flask import session
from flask_babel import Babel
from flask_babel import get_locale
from werkzeug.exceptions import NotFound
from werkzeug.routing import BaseConverter

babel = Babel()


class LangConverter(BaseConverter):
    """URL converter that validates language codes against available translations."""

    def __init__(self, map, *args, **kwargs):
        super().__init__(map, *args, **kwargs)
        self.regex = r"[a-z]{2,3}"

    def to_python(self, value):
        """Convert URL value to Python, validating against available translations.

        Raises 404 if language code is not available.
        """
        available_langs = [locale.language for locale in babel.list_translations()]

        if value not in available_langs:
            raise NotFound(
                f"Language '{value}' is not available. "
                f"Available languages: {', '.join(available_langs)}"
            )

        return value

    def to_url(self, value):
        """Convert Python value to URL."""
        return value


def setup_i18n(app):
    @app.url_defaults
    def add_language_code(endpoint, values):
        """Automatically inject language code into url_for() calls when using i18n routes."""
        if endpoint and endpoint.startswith("oauth."):
            return

        if not g.get("lang_code"):
            g.lang_code = session.get(
                "lang_code", app.config.get("BABEL_DEFAULT_LOCALE", "en")
            )

        if "lang_code" not in values:
            values.setdefault("lang_code", g.lang_code)

    @app.url_value_preprocessor
    def pull_lang_code(endpoint, values):
        """Extract language code from URL and store in g.lang_code for i18n routes."""
        if values is not None and "lang_code" in values:
            g.lang_code = values.pop("lang_code")
            session["lang_code"] = g.lang_code

    @app.before_request
    def before_request():
        g.available_languages = babel.list_translations()
        g.available_language_codes = [
            locale.language for locale in g.available_languages
        ]

    @app.context_processor
    def global_processor():
        return {
            "locale": get_locale(),
            "available_languages": getattr(g, "available_languages", []),
        }

    def locale_selector():
        # Language is always specified in URL prefix (mandatory)
        return g.get("lang_code", app.config["BABEL_DEFAULT_LOCALE"])

    babel.init_app(app, locale_selector=locale_selector)
    app.url_map.converters["lang"] = LangConverter
    app.jinja_env.policies["ext.i18n.trimmed"] = True



import os  # pragma: no cover

from dotenv import load_dotenv  # pragma: no cover

from auth_playground import create_app  # pragma: no cover


def run():  # pragma: no cover
    """Run the Auth Playground application."""
    load_dotenv()
    app = create_app()
    host = os.environ.get("FLASK_RUN_HOST", "0.0.0.0")
    port = int(os.environ.get("FLASK_RUN_PORT", "4000"))
    debug = os.environ.get("FLASK_DEBUG", "True").lower() == "true"
    app.run(host=host, port=port, debug=debug)


if __name__ == "__main__":  # pragma: no cover
    run()



import requests
from flask import Blueprint
from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import render_template
from flask import session
from flask import url_for
from flask_babel import gettext as _

import auth_playground
from auth_playground.decorators import server_config_needed
from auth_playground.forms import AuthorizationParamsForm
from auth_playground.forms import ClientConfigForm
from auth_playground.forms import DynamicRegistrationForm
from auth_playground.forms import RefreshTokenForm
from auth_playground.forms import ServerConfigForm
from auth_playground.forms import UnregisterClientForm
from auth_playground.oauth import clear_user_session
from auth_playground.session import ServerConfig

bp = Blueprint("routes", __name__)


def fetch_server_metadata(issuer_url: str, timeout: int = 10) -> tuple[dict, str]:
    """Fetch server metadata from well-known OIDC or OAuth2 endpoints."""
    oidc_url = f"{issuer_url}/.well-known/openid-configuration"
    try:
        response = requests.get(oidc_url, timeout=timeout)
        response.raise_for_status()
        return response.json(), "oidc"
    except requests.exceptions.HTTPError as e:
        if e.response.status_code != 404:
            raise

    oauth2_url = f"{issuer_url}/.well-known/oauth-authorization-server"
    response = requests.get(oauth2_url, timeout=timeout)
    response.raise_for_status()
    return response.json(), "oauth2"


def _flash_metadata_error(exception: Exception) -> None:
    """Flash user-friendly error messages for metadata loading errors."""
    if isinstance(exception, requests.exceptions.ConnectionError):
        flash(_("Cannot connect to the server. Please check the URL."), "error")
    elif isinstance(exception, requests.exceptions.Timeout):
        flash(_("Connection timeout. The server is not responding."), "error")
    elif isinstance(exception, requests.exceptions.HTTPError):
        if exception.response.status_code == 404:
            flash(
                _(
                    "This server does not support OIDC Discovery or OAuth 2.0 Authorization Server Metadata"
                ),
                "error",
            )
        else:
            flash(
                _("Server returned an error: HTTP {status_code}").format(
                    status_code=exception.response.status_code
                ),
                "error",
            )
    elif isinstance(exception, requests.RequestException):
        flash(_("Failed to connect to the server. Please check the URL."), "error")
    elif isinstance(exception, ValueError):
        flash(_("Invalid response from server"), "error")


def clear_server_session():
    """Clear all server-related data from the session."""
    clear_user_session()
    ServerConfig.clear(session)

    try:
        del session["oauth_config"]
    except KeyError:
        pass


def load_server_metadata(
    issuer_url: str | None = None,
    *,
    update_config: bool = True,
    flash_errors: bool = False,
) -> tuple[dict, str] | None:
    """Load server metadata.

    If issuer_url is None, tries to get it from environment.
    If update_config is True, updates g.server_config with metadata.
    If flash_errors is True, displays user-friendly error messages and returns None on error.
    If flash_errors is False, raises the exception.
    """
    if update_config and g.server_config and g.server_config.metadata:
        return g.server_config.metadata, g.server_config.server_type

    if not issuer_url:
        issuer_url = current_app.config.get("OAUTH_AUTH_SERVER")

    if not issuer_url:
        return None

    try:
        metadata, server_type = fetch_server_metadata(issuer_url)
    except Exception as e:
        if flash_errors:
            _flash_metadata_error(e)
            return None
        raise

    if update_config and g.server_config:
        g.server_config.metadata = metadata
        g.server_config.server_type = server_type
        g.server_config.issuer_url = issuer_url.rstrip("/")

    return metadata, server_type


@bp.route("/")
def index():
    """Redirect to the appropriate configuration step or main page."""
    server_configured = auth_playground.is_server_configured()
    client_configured = auth_playground.is_client_configured()

    if not server_configured:
        return redirect(url_for("routes.configure_server"))

    elif not client_configured:
        return redirect(url_for("routes.configure_client"))

    return redirect(url_for("routes.playground"))


@bp.route("/server", methods=["GET", "POST"])
@bp.route("/server/<path:domain>", methods=["GET", "POST"])
def configure_server(domain=None):
    """Display form to configure and validate identity provider URL."""
    if auth_playground.is_oauth_server_from_env(current_app):
        flash(
            _(
                "OAuth configuration is set via environment variables and cannot be changed"
            ),
            "warning",
        )
        return redirect(url_for("routes.playground"))

    form = ServerConfigForm()

    if domain:
        if not domain.startswith(("http://", "https://")):
            domain = f"https://{domain}"
        issuer_url = domain.rstrip("/")

    elif not form.validate_on_submit():
        if g.server_config and g.server_config.issuer_url:
            flash(_("You can now configure a different identity provider"), "info")
        clear_server_session()
        g.server_config = ServerConfig()
        return render_template("configure_server.html", form=form)

    else:
        issuer_url = form.issuer_url.data.rstrip("/")

    result = load_server_metadata(issuer_url, update_config=False, flash_errors=True)
    if not result:
        return render_template("configure_server.html", form=form)
    metadata, server_type = result

    g.server_config = ServerConfig(
        metadata=metadata,
        issuer_url=issuer_url,
        server_type=server_type,
    )

    flash(_("Server metadata successfully loaded"), "success")

    return redirect(url_for("routes.configure_client"))


@bp.route("/client", methods=["GET", "POST"])
@server_config_needed(client_needed=False)
def configure_client():
    """Display options to configure OAuth client credentials."""
    if auth_playground.is_oauth_client_from_env(current_app):
        flash(
            _("OAuth client is set via environment variables and cannot be changed"),
            "warning",
        )
        return redirect(url_for("routes.playground"))

    client_form = ClientConfigForm()
    dynamic_registration_form = DynamicRegistrationForm()

    if not client_form.validate_on_submit():
        return render_template(
            "configure_client.html",
            client_form=client_form,
            dynamic_registration_form=dynamic_registration_form,
        )

    auth_playground.setup_oauth_runtime(
        current_app,
        client_form.client_id.data,
        client_form.client_secret.data,
        g.server_config.issuer_url,
    )
    flash(_("OAuth configuration completed successfully"), "success")
    return redirect(url_for("routes.playground"))


@bp.route("/playground")
@server_config_needed
def playground():
    """Display the main playground page with OAuth 2.0 demonstration controls."""
    refresh_form = RefreshTokenForm()
    unregister_form = UnregisterClientForm()
    auth_params_form = AuthorizationParamsForm()

    return render_template(
        "playground.html",
        refresh_form=refresh_form,
        unregister_form=unregister_form,
        auth_params_form=auth_params_form,
    )


@bp.route("/specs")
@server_config_needed(client_needed=False)
def specs():
    """Display server specifications."""
    if not g.server_config.specs:
        flash(_("No server configured"), "warning")
        return redirect(url_for("routes.configure_server"))

    server_specs = g.server_config.specs
    supported_specs = [
        (
            spec,
            server_specs.get_spec_display_name(spec),
            server_specs.get_spec_url(spec),
        )
        for spec in server_specs.get_supported_specs()
    ]
    unsupported_specs = [
        (
            spec,
            server_specs.get_spec_display_name(spec),
            server_specs.get_spec_url(spec),
        )
        for spec in server_specs.get_unsupported_specs()
    ]
    unknown_specs = [
        (
            spec,
            server_specs.get_spec_display_name(spec),
            server_specs.get_spec_url(spec),
        )
        for spec in server_specs.get_unknown_specs()
    ]

    return render_template(
        "specs.html",
        supported_specs=supported_specs,
        unsupported_specs=unsupported_specs,
        unknown_specs=unknown_specs,
    )


@bp.route("/tos")
def tos():
    """Display the Terms of Service page."""
    return render_template("tos.html")


@bp.route("/policy")
def policy():
    """Display the Privacy Policy page."""
    return render_template("policy.html")


@bp.route("/tokens")
@server_config_needed
def tokens():
    """Display token information page."""
    return render_template("tokens.html")



from urllib.parse import urlparse

from babel import Locale
from babel import UnknownLocaleError
from flask import current_app
from flask import g
from flask_babel import lazy_gettext as _
from flask_wtf import FlaskForm
from wtforms import PasswordField
from wtforms import SelectField
from wtforms import SelectMultipleField
from wtforms import StringField
from wtforms import SubmitField
from wtforms import ValidationError
from wtforms.validators import DataRequired
from wtforms.validators import Length
from wtforms.widgets import CheckboxInput
from wtforms.widgets import ListWidget


def get_scopes_choices():
    """Get scopes choices from server metadata."""
    if g.server_config and g.server_config.metadata:
        scopes_supported = g.server_config.metadata.get("scopes_supported", [])
        return [(scope, scope) for scope in scopes_supported]
    return []


def get_scopes_default():
    """Get default scopes (all supported scopes)."""
    if g.server_config and g.server_config.metadata:
        return g.server_config.metadata.get("scopes_supported", [])
    return []


def get_prompt_choices():
    """Get prompt choices for OAuth authorization."""
    return [
        ("", _("Default")),
        ("none", _("None - No UI")),
        ("login", _("Login - Authentication page")),
        ("consent", _("Consent - Consent page")),
        ("select_account", _("Select account - Account selection page")),
        ("create", _("Create - Registration page")),
    ]


def get_ui_locales_choices():
    """Get UI locales choices from server metadata."""
    choices = [("", _("Default"))]
    if (
        g.server_config
        and g.server_config.metadata
        and (
            ui_locales_supported := g.server_config.metadata.get(
                "ui_locales_supported", []
            )
        )
    ):
        for locale_code in ui_locales_supported:
            try:
                # Parse the locale and get its display name in its own language
                locale_obj = Locale.parse(locale_code.replace("-", "_"))
                display_name = locale_obj.get_display_name(
                    locale_code.replace("-", "_")
                )
                choices.append((locale_code, display_name))

            except (UnknownLocaleError, ValueError):
                # Fallback to the locale code if parsing fails
                choices.append((locale_code, locale_code))
    return choices


class RefreshTokenForm(FlaskForm):
    """Form to refresh access token using refresh token."""

    submit = SubmitField(_("Renew tokens"))


def validate_issuer_url(form, field):
    """Validate issuer URL with relaxed rules in debug/testing mode or for localhost."""
    url = field.data

    if not url.startswith(("http://", "https://")):
        raise ValidationError(_("URL must start with http:// or https://"))

    is_debug = current_app.debug
    is_testing = current_app.testing
    parsed = urlparse(url)
    hostname = parsed.hostname or ""
    # RFC 8252 Section 8.3: HTTP is allowed for loopback interface redirect URIs
    # https://datatracker.ietf.org/doc/html/rfc8252#section-8.3
    is_localhost = hostname in ("localhost", "127.0.0.1", "::1")

    if url.startswith("http://") and not (
        is_debug or is_testing or is_localhost
    ):  # pragma: no cover
        raise ValidationError(
            _(
                "HTTP is only allowed for localhost or in debug/testing mode. Use HTTPS in production."
            )
        )


class ServerConfigForm(FlaskForm):
    """Form to configure the provider server URL."""

    issuer_url = StringField(
        _("Provider URL:"),
        validators=[
            DataRequired(message=_("Provider URL is required")),
            validate_issuer_url,
        ],
        description=_(
            "Enter the base URL of your OIDC/OAuth2 provider (e.g., https://auth.example.com)"
        ),
        render_kw={"placeholder": "https://auth.example.com", "type": "url"},
    )
    submit = SubmitField(_("Continue"))


class ClientConfigForm(FlaskForm):
    """Form to configure OAuth client credentials manually."""

    client_id = StringField(
        _("Client ID:"),
        validators=[
            DataRequired(message=_("Client ID is required")),
            Length(
                min=1,
                max=255,
                message=_("Client ID must be between 1 and 255 characters"),
            ),
        ],
        render_kw={"placeholder": "auth-playground-client"},
    )
    client_secret = PasswordField(
        _("Client secret:"),
        validators=[
            DataRequired(message=_("Client Secret is required")),
            Length(min=1, message=_("Client Secret is required")),
        ],
        render_kw={"placeholder": "******************"},
    )
    submit = SubmitField(_("Complete configuration"))


class DynamicRegistrationForm(FlaskForm):
    """Form to trigger dynamic client registration with CSRF protection."""

    initial_access_token = StringField(
        _("Initial access token:"),
        validators=[],
        render_kw={"placeholder": _("Leave empty if not required")},
    )
    submit = SubmitField(_("Register client"))


class UnregisterClientForm(FlaskForm):
    """Form to trigger client unregistration with CSRF protection."""

    submit = SubmitField(_("Unregister client"))


class AuthorizationParamsForm(FlaskForm):
    """Form to customize OAuth authorization parameters."""

    scopes = SelectMultipleField(
        _("Scopes:"),
        choices=get_scopes_choices,
        default=get_scopes_default,
        description=_("Select the scopes to request from the authorization server"),
        widget=ListWidget(prefix_label=False),
        option_widget=CheckboxInput(),
    )
    prompt = SelectField(
        _("Prompt:"),
        choices=get_prompt_choices,
        description=_("Control the authentication UI behavior"),
    )
    ui_locales = SelectField(
        _("UI Locale:"),
        choices=get_ui_locales_choices,
        description=_("Preferred language for the authentication UI"),
    )



from functools import wraps

from flask import redirect
from flask import url_for

import auth_playground


def server_config_needed(f=None, *, client_needed=True):
    """Ensure server is configured and metadata is loaded.

    If client_needed is True, also ensures OAuth client is configured.
    """

    def decorator(func):
        @wraps(func)
        def decorated_function(*args, **kwargs):
            from auth_playground.endpoints import load_server_metadata

            if not auth_playground.is_server_configured():
                return redirect(url_for("routes.configure_server"))

            load_server_metadata(flash_errors=True)

            if client_needed and not auth_playground.is_client_configured():
                return redirect(url_for("routes.configure_client"))

            return func(*args, **kwargs)

        return decorated_function

    if f is None:
        return decorator
    else:
        return decorator(f)



from __future__ import annotations

from dataclasses import dataclass

from auth_playground.specs import ServerSpecs


@dataclass
class ServerConfig:
    """Encapsulates server configuration data from session."""

    metadata: dict | None = None
    issuer_url: str | None = None
    server_type: str | None = None
    registration_access_token: str | None = None
    registration_client_uri: str | None = None

    @property
    def specs(self) -> ServerSpecs | None:
        """Get ServerSpecs for the configured server."""
        if not self.metadata:
            return None
        return ServerSpecs(self.metadata)

    @property
    def display_name(self) -> str:
        """Get display name for the server type."""
        if self.server_type == "oidc":
            return "OpenID Provider"
        else:
            return "OAuth 2.0 Authorization Server"

    def serialize(self) -> dict:
        """Serialize to dict for session storage."""
        return {
            "server_metadata": self.metadata,
            "issuer_url": self.issuer_url,
            "server_type": self.server_type,
            "registration_access_token": self.registration_access_token,
            "registration_client_uri": self.registration_client_uri,
        }

    @classmethod
    def deserialize(cls, session_data: dict) -> ServerConfig | None:
        """Deserialize from session storage."""
        metadata = session_data.get("server_metadata")
        issuer_url = session_data.get("issuer_url")
        server_type = session_data.get("server_type")
        registration_access_token = session_data.get("registration_access_token")
        registration_client_uri = session_data.get("registration_client_uri")

        if not metadata and not issuer_url:
            return None

        return cls(
            metadata=metadata,
            issuer_url=issuer_url,
            server_type=server_type,
            registration_access_token=registration_access_token,
            registration_client_uri=registration_client_uri,
        )

    def save(self, session_data: dict):
        """Save to session storage."""
        serialized = self.serialize()
        for key, value in serialized.items():
            if value is not None:
                session_data[key] = value
            else:
                session_data.pop(key, None)

    @classmethod
    def clear(cls, session_data: dict):
        """Clear from session storage."""
        session_data.pop("server_metadata", None)
        session_data.pop("issuer_url", None)
        session_data.pop("server_type", None)
        session_data.pop("registration_access_token", None)
        session_data.pop("registration_client_uri", None)
